"""
cs5win

meant to be like Zelle's window, but a bit more flexible...
"""

import random
import Queue
import threading
import socket
import time
import Tkinter
import time
import math
import sys

class Affine2dTrans:
    """ let's create our own 2d affine class """

    def __init__(self):
        """ the components currently supported """ 
        self.scalex = 1.0  # units of  pixels per worldunit (cm)
        self.scaley = 1.0  # units of  pixels per worldunit (cm)
        self.cx = 0.0      # worldunits (cm)
        self.cy = 0.0      # worldunits (cm)
        self.px = 0.0      # pixels
        self.py = 0.0      # pixels
        self.thr = 0.0     # radians, though the interface is always degrees
        self.sinthr = math.sin(0.0)   # why compute this more than once?
        self.costhr = math.cos(0.0)   # why not say it more than once?

    def transform(self, (x,y)):
        return self.transformWorldToPixel((x,y))

    def transformWorldToPixel(self, (worldx, worldy)):
        """ outpus pixel coordinates for the input world coords """
        # first, compute the offset from the world's center of rotation
        offx = worldx - self.cx
        offy = worldy - self.cy
        # next, rotate the resulting vector by theta
        rx = self.costhr * offx - self.sinthr * offy
        ry = self.sinthr * offx + self.costhr * offy
        # next, scale the result
        rx = rx * self.scalex
        ry = ry * self.scaley
        # finally, add the pixel offset
        return ( self.px + rx, self.py + ry )

    def transformPixelToWorld(self, (pixelx, pixely)):
        """ outputs world coordinates for the input pixel coords """
        # first subtract the pixel offset
        rx = pixelx - float(self.px)
        ry = pixely - float(self.py)
        # unscale the result
        rx = rx / self.scalex
        ry = ry / self.scaley
        # unrotate the vector by theta
        offx =  self.costhr * rx + self.sinthr * ry
        offy = -self.sinthr * rx + self.costhr * ry
        # add the world offset to get world coords
        return (offx + self.cx, offy + self.cy)

    def transform_scale(self, length, thd):
        """ returns the new length equivalent in pixels to  the input
            length along the thd direction (in degrees) - if x and y
            are not equal, thd will make a difference... 

            length sould be positive - a positive result will be
            returned in any case

            thd is in global world terms, not relative to the
            stored thr within this transformation """
        # first, compute the offset from the world's center of rotation
        thr = thd * 180.0 / math.pi
        offx = length * math.cos(thr)
        offy = length * math.sin(thr)
        # next, scale the result
        rx = offx * self.scalex
        ry = offy * self.scaley
        # finally, compute the magnitude
        return ( math.sqrt( rx * rx + ry * ry ) )

    def setWorldRotationCenter(self, worldcenterx, worldcentery):
        """ puts the input world point onto the input pixel point
            and make that point the center of any rotation specified """
        # I don't know if rotation should be preserved through this call...
        self.cx = worldcenterx
        self.cy = worldcentery

    def setPixelRotationCenter(self, pixelx, pixely):
        """ puts the input world point onto the input pixel point
            and make that point the center of any rotation specified """
        # I don't know if rotation should be preserved through this call...
        self.px = pixelx
        self.py = pixely

    def setScales(self, scale_x, scale_y):
        """ allows for axis flipping and nonuniform (!) scaling, though
            why you'd want that is beyond me... each of scale_x and 
            scale_y are expressed in pixels per world unit (cm) "
        """
        if scale_x == 0.0 or scale_y == 0.0:
            print "Warning - you don't want a finite world length mapping to 0 pixels!"
        self.scalex = scale_x
        self.scaley = scale_y

    def multiplyScales(self, mult_x, mult_y):
        """ change the x and y scales by multiplying by the inputs 
            scalex and scaley are in units of pixels per world unit (cm)
        """
        if mult_x == 0.0 or mult_y == 0.0:
            print "Warning - you don't want a finite world length mapping to 0 pixels!"
        self.scalex *= mult_x
        self.scaley *= mult_y

    def setRotationAngle(self, thd):
        """ set the number of degrees which the world coordinates
            are rotated (around the rotation center) before mapping to pixels """
        self.thr = thd * math.pi / 180.0
        self.costhr = math.cos(self.thr)
        self.sinthr = math.sin(self.thr)

    def deltaRotationAngle(self, thd):
        """ changes the number of degrees which the world coordinates
            are rotated (around the rotation center) before mapping to pixels """
        self.thr = self.thr + (thd * math.pi / 180.0)
        self.costhr = math.cos(self.thr)
        self.sinthr = math.sin(self.thr)


class Movable:
    def __init__(self, listofvertices, cx, cy, thd):
        """ set up a local coordinate system """
        self.typeName = 'Movable'
        self.orig_vertices = listofvertices
        self.cx = cx
        self.cy = cy
        self.thr = thd * math.pi/180.0
        self.sinthr = math.sin(self.thr)
        self.costhr = math.cos(self.thr)
        self.updateVertices()

    # this changes vertices from their movable-local coords
    # to global (world) coords
    # the transformation to pixel coords happens in redraw...
    def adjustPose(self, (x,y)):
        return ( (x*self.costhr - y*self.sinthr + self.cx, 
                  x*self.sinthr + y*self.costhr + self.cy) )

    def updateVertices(self):
        """ change vertices for a new cx, cy, thr """
        self.vertices =  map( self.adjustPose, self.orig_vertices )

    def deltaPose( self, dx, dy, dthr ):
        self.cx += dx
        self.cy += dy
        self.thr += dthr
        self.sinthr = math.sin(self.thr)
        self.costhr = math.cos(self.thr)
        self.updateVertices()

    def setPose( self, cx, cy, thr ):
        self.cx = cx
        self.cy = cy
        self.thr = thr
        self.sinthr = math.sin(self.thr)
        self.costhr = math.cos(self.thr)
        self.updateVertices()


# does this even get used?
class Polygon:
    def __init__(self, listofvertices, segIndex=-1):
        self.vertices = listofvertices
        self.id = segIndex



# here is the visualizer class...
class TkVisualizer(Tkinter.Toplevel):

    def __init__(self, root=None, h=400, w=400):

        # basic data members 
        self.root = root

        # base class constructor
        Tkinter.Toplevel.__init__(self, root)

        # starting height and width
        self._height = h
        self._width = w

        # previous Visualizer members
        self.tfm = Affine2dTrans()
        self.tfm.setWorldRotationCenter(0,0) # maps the world point 0,0 to 
        self.tfm.setPixelRotationCenter(self._width/2,self._height/2) # the center of the window

        # perhaps this should be called map or something less general
        self.world = []      # for stationary obstacles
        self.movables = {}   # for possibly moving objects with a local coordinate system

        # here are some things for displaying a graph
        self.graphPoints = []  # five-tuples x y r g b
        self.graphEdges = [] # seven-tuples x y x y r g b

        # dictionary for keys pressed
        self.keysdown = {}

        # some global window-handling stuff
        self.wm_title("Robot/World Visualizer")
        self.protocol('WM_DELETE_WINDOW',self.destroy)

        # make the whole application resizable
        self.rowconfigure(0,weight=0) # row 0 = menu bar on top
        self.rowconfigure(1,weight=1) # row 1 = canvas with vizualizations
        self.columnconfigure(0,weight=1)         # everything in one column (for now)
    
        # the canvas for general drawing
        self.canvas = Tkinter.Canvas(self, bg="white", width=self._width, height=self._height)
        self.canvas.grid(column=0, row=1, sticky='news')  
        self.canvas.bind("<Configure>", self.configcallback)

        # this will bring the focus to the main canvas...
        self.canvas.focus_set()
        self.canvas.bind("<KeyPress>", self.keypresscallback)
        self.canvas.bind("<KeyRelease>", self.keyreleasecallback)
        self.canvas.config(highlightthickness=0)

        # handle all of the button events...
        self.canvas.bind("<Button-1>", self.b1_down)
        self.canvas.bind("<B1-Motion>", self.b1_move)
        self.canvas.bind("<ButtonRelease-1>", self.b1_up)

        self.canvas.bind("<Button-2>", self.click_b2_down)
        self.canvas.bind("<B2-Motion>", self.click_b2_move)
        self.canvas.bind("<ButtonRelease-2>", self.click_b2_up)

        self.canvas.bind("<Button-3>", self.click_b3_down)
        self.canvas.bind("<B3-Motion>", self.click_b3_move)
        self.canvas.bind("<ButtonRelease-3>", self.click_b3_up)

        # the menu bar above the canvas
        self.mBar = Tkinter.Frame(self, relief=Tkinter.FLAT, borderwidth=1)
        self.mBar.grid(column=0, row=0, sticky="w")  # the position of the mBar frame in the toplevel
    
        # this makes it easy to change the menu structure
        menuData = [
            ('File',[['Reset', self.redraw],
                     ['Exit', self.destroy] ] ),
            ('View',[
                     ['trail', lambda: self.toggle("trail")],                     
                     ['body', lambda: self.toggle("body")],                 
                     ['boundingBox', lambda: self.toggle("boundingBox")],
                     ['sonar', lambda: self.toggle("sonar")],
                     ['light', lambda: self.toggle("light")],                     
                     ['lightBlocked', lambda: self.toggle("lightBlocked")], 
                    ]
             ), ]

        # this code iterates along the above structure to create the menus...
        self.menubuttonlist = []
        columnnumber = 0
        for entry in menuData:
            self.menubuttonlist.append( Tkinter.Menubutton( self.mBar, text=entry[0] ) )
            thismenubutton = self.menubuttonlist[-1]
            thismenubutton.grid(column=columnnumber, row=0, sticky="w") # position to the left
            columnnumber += 1
            # one Menu (the whole thing that pulls down)
            thismenubutton.menu = Tkinter.Menu( self.menubuttonlist[-1], tearoff=1 )
            thismenubutton["menu"] = thismenubutton.menu
            # populate the pull-down Menu
            for item in entry[1]:
                thismenubutton.menu.add_command( label=item[0], command=item[1] )

        # let's add a label to the menu bar
        self.keysDownLabelText = "Keys Down: " # space after the colon is important for later processing
        self.keysDownLabel = Tkinter.Label( self.mBar, text=self.keysDownLabelText, fg="green" )
        self.keysDownLabel.grid( row=0, column=columnnumber, sticky="ew" )
        columnnumber += 1


        # we'll run it from both for now...
        # run at no more than 10 hertz for right now (100 milliseconds)
        self.idleTime = 20 # milliseconds
        self.idlerID = self.after(self.idleTime, self.idleTasks) 

	# end of the __init__ constructor

    def idleTasks(self):
        #self.processIncoming()
        # need to reregister this idle handler
        self.idlerID = self.after(self.idleTime,self.idleTasks)
        #print 'hi there'

    def toggle(self, key):
        for r in self.movables:
            if r.subscribed == 0: continue
            if r.display[key] == 1:
                r.display[key] = 0
            else:
                r.display[key] = 1
            r._last_pose = (-1, -1, -1)


    # so we can start from elsewhere...
    def startmainloop(self):
        self.root.mainloop()

    def configcallback(self, event):
        #print "***** config! *****"
        #print "dir is", dir(event)
        #print 'char, delta, height, keycode are', event.char, event.delta, event.height, event.keycode
        #print 'keysym, keysym_num, num, send_event', event.keysym, event.keysym_num, event.num, event.send_event
        #print 'serial, state, time, type', event.serial, event.state, event.time, event.type
        #print 'widget, width, x, x_root, y, y_root', event.widget, event.width, event.x, event.x_root, event.y, event.y_root
        #print "***** ******* *****"
        self._height = event.height 
        self._width = event.width
        # so that the world rotation center maps to the center of the window...
        self.tfm.setPixelRotationCenter(self._width/2,self._height/2) 
        self.redraw()
    
    def keypresscallback(self, event):
        c = event.char
        #print 'keysym is', event.keysym

        # add an entry to our keysdown dictionary
        self.keysdown[event.keysym] = 1

        # change the keys down label...
        # simply add the new char to the string (at the front)
        # but only if it's not a space and not already there...
        s = self.keysDownLabelText
        if c != ' ' and s.find(c, s.find(":")) == -1:
            self.keysDownLabelText = s.replace(": ", ": "+c, 1) 
            self.keysDownLabel.config( text=self.keysDownLabelText, fg = "red" )

        if c == 'q':
            self.destroy() # cleaner!
        elif c == 'p':  # pause
            # this should ignore incoming messages
            # until further notice
            self.ignoreRobotUpdates = not self.ignoreRobotUpdates
        elif c == 'R': # show image from robot
            print 'taking an image from the robot'
            self.retqueue.put("showRobotImage")
        elif c == 'P':
            # is there a camera?
            if self.movables.has_key("generic_camera"):
                x = self.movables["generic_camera"].cx
                y = self.movables["generic_camera"].cy
                thr = self.movables["generic_camera"].thr
                print 'pose adjusting with cam at', x, y, thr*180.0/math.pi
                self.retqueue.put("poseAdjust_%f_%f_%f" % (x,y,thr))
            else:
                print 'plain pose adjust'
                self.retqueue.put("poseAdjust")
        elif c == 'a':
            self.tfm.deltaRotationAngle(10)
            self.redraw()
        elif c == 'A':
            if self.movables.has_key("er2"):
                self.movables["er2"].deltaPose(10,0,0.0)
            self.redraw()
        elif c == 'b':
            self.canvas.move("cool",20,0)
            self.redraw()
        elif c == '?':
            # way to ask about things on the fly!
            #print 'dir is', dir(tkv.canvas)
            #print 'help is', help(tkv.canvas.create_polygon)
            pass
        
        #  bad interaction with press-and-hold keys... need to
        #  put this in keyrelease instead of key press!!
        #elif len(c) == 1 and ord('A') <= ord(c) <= ord('z'):
        #    print 'sending char:', c
        #    self.retqueue.put("char:%s" % (c,))
        else:
            #print 'key pressed is', event.char
            pass

        # char is the empty string for special (non-char) keys
        #print 'keysym is', event.keysym # things like 'Shift_R' for right shift key, etc.

    def keyreleasecallback(self, event):
        k = event.keysym
        # print "key RELEASE of", k
        self.keysdown[event.keysym] = 0

        # remove the key from the "keys down" display
        c = event.char
        s = self.keysDownLabelText
        if c != ' ' and s.find(c, s.find(":")) != -1:
            # it's there... get everything after the semicolon
            s_tail = s[s.find(": ")+2:]
            s_tail = s_tail.replace(c,'')   # replace with nothing
            fgcolor = "red"
            if len(s_tail) == 0:    fgcolor = "green"   # set to green if nothing's left
            self.keysDownLabelText = "Keys Down: " + s_tail
            self.keysDownLabel.config( text=self.keysDownLabelText, fg = fgcolor )

    def destroy(self):
        self.running = 0 # for thread
        self.done = 1 # stop processing requests, if handling
        self.quit = 1 # stop accept/bind toplevel
        self.root.quit() # kill the gui


    def b1_down(self, event):
        if self.keysdown.has_key("Shift_L") and self.keysdown["Shift_L"] == 1:
            # print "b1 down with shift"
            self.click_b3_down(event)

        elif self.keysdown.has_key("c") and self.keysdown["c"] == 1:
            self.startx = event.x
            self.starty = event.y
            

    def b1_move(self,event):
        if self.keysdown.has_key("Shift_L") and self.keysdown["Shift_L"] == 1:
            # print "b1 move with shift"
            self.click_b3_move(event)

        elif self.keysdown.has_key("c") and self.keysdown["c"] == 1:
            self.endx = event.x
            self.endy = event.y
            # create a line - only in pixel space for now...
            self.canvas.delete('temporary_line')
            id = self.canvas.create_line(self.startx,self.starty,self.endx,self.endy,
                                         tag='temporary_line', fill='black')


    def b1_up(self,event):
        #print "up event at pixel coordinate", event.x, event.y
        #print "up event.keysym is", event.keysym
        if self.keysdown.has_key("Shift_L") and self.keysdown["Shift_L"] == 1:
            # print "b1 up with shift"
            self.click_b3_up(event)
            
        elif self.keysdown.has_key("Control_L") and self.keysdown["Control_L"] == 1:
            (wx,wy) = self.tfm.transformPixelToWorld((event.x, event.y))
            self.tfm.setWorldRotationCenter(wx, wy)
            self.redraw()
            
        elif self.keysdown.has_key("c") and self.keysdown["c"] == 1:
            self.endx = event.x
            self.endy = event.y
            # delete old line
            self.canvas.delete('temporary_line')
            wx, wy = self.tfm.transformPixelToWorld( (self.startx,self.starty) )
            # print "world coordinates are (wx,wy) = ", (wx, wy)
            # compute the theta in world space!
            ex, ey = self.tfm.transformPixelToWorld( (self.endx,self.endy) )
            thr = math.atan2( ey-wy, ex-wx)
            thd = (180.0/math.pi) * thr
            # create a camera!
            self.addMovable( Camera( [10.0, 8.0], wx, wy, thd, color="black" ) )
            self.redraw()
            self.retqueue.put("cpose_%f_%f_%f" % (wx,wy,thr))
            
        else:  # if no modifier, we will simply print the clicked pixel's window and world coords
            print "pixel coordinates are (px,py) = ", event.x, event.y
            print "world coordinates are (wx,wy) = ", self.tfm.transformPixelToWorld( (event.x,event.y) )


    def click_b2_down(self, event):
        pass

    def click_b2_move(self, event):
        pass

    def click_b2_up(self, event):
        pass


    def length(self, (x1, y1), (x2, y2)):
        return math.sqrt( (x1-x2)**2 + (y1-y2)**2 )

    def click_b3_down(self, event):
        self.click_start = event.x, event.y
        self.click_b3_move(event)

    def click_b3_move(self, event):
        self.canvas.delete('arrow')
        x, y = self._width/2.0, self._height/2.0
        radius = self.length((x,y),(event.x,event.y)) 
        self.canvas.create_oval(x - radius, y - radius, x + radius, y + radius, tag="arrow", outline="purple")

    def click_b3_up(self, event):
        """ scaling... """
        self.canvas.delete('arrow')
        x, y = self._width/2.0, self._height/2.0
        radius_stop = self.length((x,y),(event.x,event.y)) 
        radius_start = self.length((x,y),self.click_start)
        scaleMultiplier = radius_stop/radius_start
        self.tfm.multiplyScales(scaleMultiplier, scaleMultiplier)
        self.redraw()


    def redraw(self):
        # print "Redrawing things..."
        self.canvas.delete('all')
        for pt in self.graphPoints:
            x, y, r, g, b = pt
            (cx,cy) = self.tfm.transform((x,y))
            # radius is 15 right now
            radius = self.tfm.transform_scale(15.0,0.0)
            self.canvas.create_oval(cx-radius,cy-radius,cx+radius,cy+radius,fill=("#%02x%02x%02x" % (r,g,b)),outline="black")
                
        for edge in self.graphEdges:
                x1, y1, x2, y2, r, g, b = edge
                (x1,y1) = self.tfm.transform((x1,y1))
                (x2,y2) = self.tfm.transform((x2,y2))
                id = self.canvas.create_line(x1,y1,x2,y2, fill=("#%02x%02x%02x" % (r,g,b)), width=2)

        for shape in self.world:

            if shape[0] == "circle":
                #print "circle display!"
                name, cx, cy, radius, color, usertag = shape
                (cx,cy) = self.tfm.transform((cx,cy))
                radius = self.tfm.transform_scale(radius,0.0)
                self.canvas.create_oval(cx-radius,cy-radius,cx+radius,cy+radius,tag=usertag,fill=color,outline="black")

            if shape[0] == "segment":
                #print '*',
                name, x1, y1, x2, y2, color, usertag = shape
                (x1,y1) = self.tfm.transform((x1,y1))
                (x2,y2) = self.tfm.transform((x2,y2))
                id = self.canvas.create_line(x1,y1,x2,y2, tag=usertag, fill=color, width=2)

            if shape[0] == "particles":
                name, particleList, color, usertag = shape
                for particle in particleList:
                    x1, y1, x2, y2 = particle
                    (x1,y1) = self.tfm.transform((x1,y1))
                    (x2,y2) = self.tfm.transform((x2,y2))
                    id = self.canvas.create_line(x1,y1,x2,y2, tag=usertag, fill=color, arrow=Tkinter.LAST)

            if shape[0] == "arrow":
                name, x1, y1, x2, y2, color, usertag = shape
                (x1,y1) = self.tfm.transform((x1,y1))
                (x2,y2) = self.tfm.transform((x2,y2))
                id = self.canvas.create_line(x1,y1,x2,y2, tag=usertag, width=2, fill=color)
                #id = self.canvas.create_line(x1,y1,x2,y2, tag=usertag, width=2, fill=color, arrow=Tkinter.LAST)


            if shape[0] == "poly":
                name, listofvertices, color, usertag = shape
                transformed_vertices = map( self.tfm.transform, listofvertices )
                id = self.canvas.create_polygon(transformed_vertices, tag=usertag, fill=color, outline="black")
                # handle id...

        for (key,item) in self.movables.iteritems():

            itemname = getattr(item, 'typeName', None)
            #print 'itemname is', itemname

            if itemname == 'Robot':
                id = item.draw(self)
                self.canvas.lower(id)
                #transformed_vertices = map( self.tfm.transform, item.vertices )
                #id = self.canvas.create_polygon(transformed_vertices, fill=item.color, outline="black")
                

            if itemname == 'Camera':
                transformed_vertices = map( self.tfm.transform, item.vertices )
                #print "item.vertices are", item.vertices
                #print "transformed_vertices are", transformed_vertices
                id = self.canvas.create_line(transformed_vertices[:2], fill=item.color, width=3)
                id = self.canvas.create_line(transformed_vertices[2:], fill=item.color, width=3)


    def addSegment(self, x1, y1, x2, y2, color="black", tag=None):
        self.world.append( ("segment", x1, y1, x2, y2, color, tag) )
        self.redraw()

    def graphClear(self):
        self.graphEdges = []
        self.graphPoints = []
        self.redraw()

    def addArrow(self, listofvertices, color="black", tag=None):
        (x1,y1) = listofvertices[0]
        (x2,y2) = listofvertices[1]
        self.world.append( ("arrow", x1, y1, x2, y2, color, tag) )
        #self.redraw()

    def addParticles(self, listofparticlesegments, color="black", tag="particles"):
        # remove old particles
        index = 0
        partIndex = -1
        for shape in self.world:
            if shape[0] == "particles":
                partIndex = index
            index += 1
        if partIndex > -1:
            self.world = self.world[0:partIndex] + self.world[partIndex+1:]
        # add new ones
        self.world.append( ("particles", listofparticlesegments, color, tag) )

    def addCircle(self, cx, cy, radius, color="white", tag=None):
        #Visualizer.addCircle(self, cx, cy, radius, color)
        self.world.append( ("circle", cx, cy, radius, color, tag) )
        #self.redraw()

    def addMovable(self, m):
        # check m's name - if already there, replace
        self.movables[m.name] = m
        #self.redraw()

    def addPoly(self, listofvertices, color="white", tag=None):
        self.world.append( ("poly", listofvertices, color, tag) )
        #self.redraw()



class RangeSensor:
    def __init__(self, name, geometry, arc, maxRange, noise = 0.0):
        self.type = name
        # geometry = (x, y, a) origin in meters and radians
        self.geometry = geometry
        self.arc = arc
        self.maxRange = maxRange
        self.noise = noise
        self.groups = {}
        self.scan = [0] * len(geometry) # for data

class Light:
    def __init__(self, x, y, brightness, color="yellow"):
        self.x = x
        self.y = y
        self.brightness = brightness
        self.color = color
        self._xyb = x, y, brightness # original settings for reset
        self.rgb = colorMap[color]
        self.type = "fixed"

class ThreadedClient:
    """
    Launch the main part of the GUI and the worker thread. periodicCall and
    endApplication could reside in the GUI part, but putting them here
    means that you have all the thread controls in a single place.
    """
    def __init__(self, master, PORT, HOST, MAP):
        """
        Start the GUI and the asynchronous threads. We are in the main
        (original) thread of the application, which will later be used by
        the GUI. We spawn a new thread for the worker.
        """
        self.master = master
        self.PORT = PORT
        self.HOST = HOST
        self.MAP = MAP

        # Create the queues
        self.queue = Queue.Queue()    # for messages coming into the visualizer
        self.retqueue = Queue.Queue() # for return messages

        # Set up the GUI part

        # here are some application-specific defaults,
        # based on the PORT number

        if self.PORT == 5008: # the robot wandering window
            self.vis = TkVisualizer(master,h=400,w=400,\
                            q=self.queue,rq=self.retqueue)
            self.vis.tfm.setScales(1,-1)
            self.vis.tfm.setWorldRotationCenter(0,0)
            robot = ER2( 0,0,0.0, color="blue", name="er2")
            #self.vis.addMovable( robot )
            realrobot = ER2( 0,0,0.0, color="green", name="realer2")
            #self.vis.addMovable( realrobot )

#	    # OK - let's get the convergence stuff here...
#            self.vis.loadMap('mapColors.txt')
#	    D1 = (105,39)
#	    D2 = (12,-14)
#	    D3 = (14,12)
#	    D4 = (-39,105)
#            for i in range(0,600):
#	    #for x in range( -190, 205, 30 ):
#	    #	for y in range( -190, 205, 30 ):
#                    x = random.uniform( -195,195 )
#                    y = random.uniform( -195, 195 )
#                    if -93 + 0.466*x < y < 107 + 0.466*x:
#                        self.vis.addArrow( [(x,y),D1] , color="green")
#                    if 107 + 0.466*x < y < 200 + 0.466*x:
#                        self.vis.addArrow( [(x,y),D3] , color="blue")
#                    if 200 + 0.466*x < y:
#                        self.vis.addArrow( [(x,y),D4] , color="black")
#                    if y < -93 + 0.466*x:
#                        self.vis.addArrow( [(x,y),D2] , color="cyan")
#
        elif self.PORT == 5009:  # the image window
            self.vis = TkVisualizer(root,h=50,w=400,\
                            q=self.queue,rq=self.retqueue)
            self.vis.tfm.setScales( -100, 1 ) # pixels per worldunit
            self.vis.tfm.setWorldRotationCenter(0,0)

        elif self.PORT == 5010:  # the image window
            self.vis = TkVisualizer(root,h=50,w=400,\
                            q=self.queue,rq=self.retqueue)
            self.vis.tfm.setScales( -100, 1 ) # pixels per worldunit
            self.vis.tfm.setWorldRotationCenter(0,0)


        if self.MAP != None:
            self.vis.loadMap(self.MAP)

        # Set up the thread to do asynchronous I/O
        # More can be made if necessary
        self.running = 1
        self.thread1 = threading.Thread(target=self.workerThread1)
        self.thread1.setDaemon(True)
        self.thread1.start()

        # Start the periodic call in the GUI to check if the queue contains
        # anything
        #self.periodicCall()

    def periodicCall(self):
        """
        Check every 100 ms if there is something new in the queue.
        """
        print "periodic call"
        self.vis.processIncoming()
        if not self.running:
            # This is the brutal stop of the system. You may want to do
            # some cleanup before actually shutting it down.
            import sys
            sys.exit(1)
        self.master.after(100, self.periodicCall)

    def workerThread1(self):
        """
        This is where we handle the asynchronous I/O. For example, it may be
        a 'select()'.
        One important thing to remember is that the thread has to yield
        control.
        """
        HOST = self.HOST
        PORT = self.PORT
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((HOST, PORT))
        s.listen(1)
        # let's see if this works
        s.settimeout(0.1)

        while True:  
            # this is our thread's loop
            print 'Listening for a connection...'
            networking = 0
            
            while networking == 0:
                try:
                    # this will block for a little bit of time...
                    # see the timeout parameter above
                    conn, addr = s.accept()
                    networking = 1
                    print 'Connected by', addr
                except socket.error:  # that is, no one connected...
                    networking = 0
                    time.sleep(0.1)
                    #print 'Not accepted'
    
            # we won't really need "running" when we just set this thread
            # to that status where it doesn't keep the process alive...
            # need to look that up some time...
            while networking == 1 and self.running == 1 and self.vis.running == 1:
                #
                try:
                    data = conn.recv(4096)  # used to be 1024 should be maxMessageSize
                    if len(data) >= 1:
                        # question: how does the timeout affect conn?
                        # Here WE SEND BACK RETURN MESSAGES, if any
                        # otherwise, we send back 'Hi 42'
    
                        # note that this sends back one message per
                        # message sent (at most)... this means
                        # that it would be bad if the visualizer 
                        # generates more messages than the client
                        # is generating -> would create a backlog in 
                        # retqueue
                        try:
                            msg = str(self.retqueue.get(block=False))
                            print 'Sending back the msg', msg
                        except Queue.Empty:
                            msg = 'hi 42'
    
                        conn.send(msg)  
    
                        self.queue.put(data)
                        if (data == 'q'):
                            networking = 0
                            conn.close()
                        time.sleep(0.01)
                    else:
                        networking = 0
                        self.queue.put('reset')  # reset the robot - this is specific
                        # to a specific visualizer and so should really be handled elsewhere...
                except: # no data this time...
                    pass

    def endApplication(self):
        self.running = 0
    

if __name__ == "__main__":

    import sys
    # sys.argv[0] is 'Visualizer.py'
    # sys.argv[1] should be the PORT number ~5008
    # sys.argv[2] should be the HOST (127.0.0.1)
    argc = len(sys.argv)
    PORT = 5008
    HOST = '127.0.0.1'
    MAP = None
    if argc <= 1:
        pass
    elif argc >= 2:
        try:
            PORT = int(sys.argv[1])
        except ValueError:
            MAP = sys.argv[1]
            # should check for this map existing...
            PORT = 5008

        # doesn't let us specify the map and the port!
        if argc >= 3:
            HOST = sys.argv[2]
    
    root = Tkinter.Tk()
    vis = TkVisualizer(root,h=400,w=400)
    vis.tfm.setScales(1,-1)
    vis.tfm.setWorldRotationCenter(0,0)
    vis.addSegment(-10,-10,20,10)

    root.withdraw()
    root.mainloop()
