# example roomba brain
   
from pyrobot.brain import Brain  
import time
import random

# some state constants and their action
# a state is a pair (stateName, duration, default)
#    stateName - next state to enter
#    duration - the time for it (0 == until next event)
#    default - IF duration is reached, go to this state
velocities = []
FORWARD = 0
velocities.append( (0.3,0.0) )
REVERSE = 1
velocities.append( (-0.3,0.0) )
RIGHT_TURN = 2
velocities.append( (0.0,-0.3) )
LEFT_TURN = 3
velocities.append( (0.0,0.3) )
STOP = 4
velocities.append( (0.0,0.0) )

def stateToString(stateTriple):
    s = stateName(stateTriple[0])
    if len(stateTriple) > 1: s += '  ' + str(stateTriple[1])
    if len(stateTriple) > 2: s += '  ' + stateName(stateTriple[2])
    return s

def stateName(state):
    if state == FORWARD: return "FORWARD"
    if state == REVERSE: return "REVERSE"
    if state == LEFT_TURN: return "LEFT_TURN"
    if state == RIGHT_TURN: return "RIGHT_TURN"
    if state == STOP: return "STOP"
    return "State #" + str(state)

def prettyPrint(L):
    print '[',
    for x in L:
        print ('%6.2f' % (x,) ),
    print ' ]'

# trying things out
class ARoombaBrain1(Brain):  

    # some state constants and their action

    # constructor
    def __init__(self,name,engine):
        Brain.__init__(self,name,engine)
        self.maxrange = 0.35 - 0.01  # meters - the first number should match the world
        self.state = (FORWARD, 0.0, FORWARD)
        self.oldvels = (0.0,0.0)
        self.lasttime = time.clock()
        self.laststeptime = self.lasttime

    # state-transitions decide here
    def setState(self,L,R):
       """ based on the two bump sensors """
       # timed out?
       if len(self.state) > 1 and self.state[1] > 0.0 and time.clock() - self.lasttime > self.state[1]:
           self.state = self.state[2:]
           if len(self.state) < 1:
               self.state = (FORWARD, 0.0)
           self.lasttime = time.clock()  # need to reset the clock with a new state

       # don't really want an elif here...
       if L or R:
           #nextTurn = random.choice([RIGHT_TURN, LEFT_TURN])
           nextTurn = LEFT_TURN
           nextDuration = random.uniform(0.5,2.0)
           self.state = (REVERSE, 0.25, nextTurn, nextDuration)
           self.lasttime = time.clock()  # need to reset the clock with a new state

      
    # must have a step method
    def step(self):  
        """ this is the method used by pyro to update the robot """
        # how much time has passed since our last step?
        deltasteptime = time.clock() - self.laststeptime
        self.laststeptime += deltasteptime

        # check our bumpers
        leftBump, rightBump = self.computeBumperPresses()
        if leftBump: print 'Left bumper pressed!'
        if rightBump: print 'Right bumper pressed!'

        # choose a state 
        # note that this will check the time, too
        self.setState(leftBump, rightBump)

        # and set its velocities
        vels = velocities[self.state[0]]
        if vels != self.oldvels or deltasteptime > 0.5:  
            # move if we haven't for a while
            print 'state is', stateToString(self.state)
            self.robot.move( vels[0], vels[1] )
            self.oldvels = vels

    # roomba sensors
    def computeBumperPresses(self):
        """ is the bumper pressed? """
        #print 'range values are', prettyPrint(self.robot.sonar_0)
        ranges = self.robot.sonar_0
        leftBump = min(ranges[0:4]) < self.maxrange
        rightBump = min(ranges[3:]) < self.maxrange
        return (leftBump, rightBump)




# some extra stuff for pyro
def INIT(engine):  
    assert (engine.robot.requires("range-sensor") and
            engine.robot.requires("continuous-movement"))
    return ARoombaBrain1('ARoombaBrain1', engine)  

