"""
A PyrobotSimulator world. A room with one obstacle and
a small inner room.

(c) 2005, PyroRobotics.org. Licensed under the GNU GPL.
"""

from pyrobot.simulators.pysim import *

def INIT():
    # (width, height), (offset x, offset y), scale:
    sim = TkSimulator((446,491),(21,451),80.517190)

    # x1, y1, x2, y2 in meters:
    sim.addBox(0, 0, 5, 5)
    sim.addBox(0, 4, 1, 5, "blue", wallcolor="blue")
    sim.addBox(2.5, 0, 2.6, 2.5, "green", wallcolor="green")
    sim.addBox(2.5, 2.5, 3.9, 2.6, "green", wallcolor="green")

    # port, name, x, y, th, bounding Xs, bounding Ys, color
    # Note: for circular robots, the bounding box is not really correct!
    #       hence we will simply make it a bit too small and count on
    #       our range sensors to keep us away from obstacles... .
    # (optional TK color name):
    sim.addRobot(60000, TkRoomba("RedRoomba",
                                  .5, 2.5, 0.00,
                                  ((.20, .20, -.20, -.20),
                                   (.20, -.20, -.20, .20))))  
    # add some sensors:
    # 
    sim.robots[0].addDevice(RoombaBumpers(0.35))
    #
    # change the maxrange with RoombaBumpers(8.0) (8 meters) default = 0.35
    # or change to Pioneer16Sonars() or PioneerFrontSonars()

    # do we want to use the real robot or the simulated one?
    sim.robots[0].realRobot(False)

    # other devices - camera does not work in windows
    #
    #sim.robots[0].addDevice(PioneerFrontLightSensors())
    # x, y relative to body center (beyond bounding box):
    #sim.robots[0].addDevice(BulbDevice(0.226, 0)) # pose x, pose y
    # width, height, startAngle, stopAngle, pose x, pose y, pose thr:
    #sim.robots[0].addDevice(Camera(60, 40, 60 * math.pi/180, -60 * math.pi/180, 0, 0, 0))

    return sim
