# 
# stateslate.py
#
# A GUI and server for the roomba (Erdos library)
#
# 5/1/2006 Z Dodds
#          dodds@cs.hmc.edu
#

import roomba
import time
import math
import random
import Queue

from Tkinter import * 
from robotCanvas import *

class StateSlate(Frame): 
    """ StateSlate is the class representing the GUI
        for the roomba library in roomba.py

        It works with ThreadedServer to provide socket-based
        access to the roomba.
    """

    def __init__(self, master=None):
        """ StateSlate constructor delegates component 
            creation to createWidgets
        """
        # I don't know where Tkinter.Tk()is being called... ?
        Frame.__init__(self, master) 
        self.grid(sticky=N+S+E+W)  
        # The above line is important although I have no idea what it does...

        # robot stuff
        self.roomba = None
        self.sf = None     # the last-obtained (frame-1) sensor frame
        self.sf2 = None    # the last-obtained frame-2 sensor frame
        self.sf3 = None    # the last-obtained frame-3 sensor frame
	
        # global formatting options
        self.option_add("*width", 15)
        self.option_add("*height", 2)
        self.createWidgets()

        # reset canvas's tranformation
        self.canvas1.reconfig()

        # communications with the server
        self.qToGUI = None
        self.qFromGUI = None
        self.threadedServerObject = None

	# idle tasks (including handling the server's messages)
	self.idleTime = 1000      # milliseconds
	self.idlerID = self.after(self.idleTime, self.idleTasks) 

    def idleTasks(self):
	""" idleTasks gets called every so often to handle
	    messages that have been put on qToGUI by the server
            this focuses on scheduling and queue management;
            handleMessage keeps all of the message-specific code
	"""
        ### print 'in idleTasks'

        # Let's check the robot's status every so often, if the connection
        # is open... and we have set it to continuously update
        if self.roomba != None and self.continuousupdatingv.get() == 1:
            # update appropriate section of the robot...
            # bump sensors every 1/20th of a second,
            # user interface every 1/4th of a second,
            # electrical data every 5 seconds or so...

            # update all for now (every 1 second)
            self.updateAll()

	# handle queues and messages
        msg = ''
        while self.qToGUI.qsize():
            try:
                msg = str(self.qToGUI.get(block=False))
		# ignore blank messages
		if len(msg) == 0: continue
                else: self.handleMessage(msg) # otherwise be sure to handle it!
            except Queue.Empty:
                pass
        # we really should collect all of the messages of a certain
	# type and then handle one of each...

        # reschedule idleTasks for a little while later
	self.idlerID = self.after(self.idleTime,self.idleTasks)

    def handleMessage(self, msg):
        """ handleMessage keeps all of the message-specific code
            in one place (instead of cluttering up idleTasks,
            which instead focuses on scheduling and queue management)
        """
        print 'Handling message', msg

    def setQueues(self, qToGUI, qFromGUI, server):
        """ sets the message queues by which the
            socket server and the GUI share information
        """
        self.qToGUI = qToGUI
        self.qFromGUI = qFromGUI
        self.threadedServerObject = server

    def createWidgets(self):
        """ a large method to create all of the pieces of the GUI
        """
#        top = self.winfo_toplevel()
#        top.rowconfigure(0,weight=1)
#        top.rowconfigure(1,weight=1)
#        top.columnconfigure(0,weight=1)
#        top.columnconfigure(1,weight=1)
#        self.rowconfigure(0,weight=1)
#        self.rowconfigure(1,weight=1)
#        self.columnconfigure(0,weight=1)
#        self.columnconfigure(1,weight=1)

        # exit and light up
        self.continuousupdatingv = IntVar()

        self.updatingButton = Checkbutton ( self, text="Updating", indicatoron=0,command=self.continuousUpdating, variable=self.continuousupdatingv ) 
        self.updatingButton.grid(row=0, column=0, sticky=N+S+E+W)
        # updaters
        self.updateButton = Button ( self, text="Update All", command=self.updateAll ) 
        self.updateButton.grid(row=0, column=1, sticky=N+S+E+W)
        self.updateButton = Button ( self, text="Frame 1", command=self.update1 ) 
        self.updateButton.grid(row=0, column=2, sticky=N+S+E+W)
        self.updateButton = Button ( self, text="Frame 2", command=self.update2 ) 
        self.updateButton.grid(row=0, column=3, sticky=N+S+E+W)
        self.updateButton = Button ( self, text="Frame 3", command=self.update3 ) 
        self.updateButton.grid(row=0, column=4, sticky=N+S+E+W)
        # safe vs. full mode
        self.stopButton = Button ( self, text="STOP", fg="red", command=self.stop ) 
        self.stopButton.grid(row=0, column=5, sticky=N+S+E+W)

        # bumper-press labels
        self.leftBumperLabel = Label(self, text="Left Bumper")
        self.leftBumperLabel.grid(row=2,column=0, sticky=N+S+E+W)
        self.rightBumperLabel = Label(self, text="Right Bumper")
        self.rightBumperLabel.grid(row=2,column=1, sticky=N+S+E+W)
        self.leftCliffLabel = Label(self, text="Left Cliff")
        self.leftCliffLabel.grid(row=2,column=2, sticky=N+S+E+W)
        self.leftCentCliffLabel = Label(self, text="LFnt Cliff")
        self.leftCentCliffLabel.grid(row=2,column=3, sticky=N+S+E+W)
        self.rightCentCliffLabel = Label(self, text="RFnt Cliff")
        self.rightCentCliffLabel.grid(row=2,column=4, sticky=N+S+E+W)
        self.rightCliffLabel = Label(self, text="Right Cliff")
        self.rightCliffLabel.grid(row=2,column=5, sticky=N+S+E+W)

        # other labels
        self.wallLabel = Label(self, text="Right Wall")
        self.wallLabel.grid(row=3,column=0, sticky=N+S+E+W)
        self.virtualWallLabel = Label(self, text="Virtual Wall")
        self.virtualWallLabel.grid(row=3,column=1, sticky=N+S+E+W)
        self.leftWheelDropLabel = Label(self, text="Left Wheel Drop")
        self.leftWheelDropLabel.grid(row=3,column=2, sticky=N+S+E+W)
        self.casterDropLabel = Label(self, text="Caster Wheel Drop")
        self.casterDropLabel.grid(row=3,column=3, sticky=N+S+E+W)
        self.rightWheelDropLabel = Label(self, text="Right Wheel Drop")
        self.rightWheelDropLabel.grid(row=3,column=4, sticky=N+S+E+W)
	self.overcurrentsAndDDLabel = Label(self, text="OC/DD ok")
        self.overcurrentsAndDDLabel.grid(row=3,column=5, sticky=N+S+E+W)


        # motor indicators
        self.vacuumIntVar = IntVar()
        self.vacuumMotor = Checkbutton( self, indicatoron=0, 
                        bd=2, variable = self.vacuumIntVar, text="Vacuum",
                        command = self.setVacuumMotor )
        self.vacuumMotor.grid(row=4, column=0, sticky=N+S+E+W )
        self.sideIntVar = IntVar()
        self.sideMotor = Checkbutton( self, indicatoron=0, 
                        bd=2, variable = self.sideIntVar, text="Side Brush",
                        command = self.setSideMotor )
        self.sideMotor.grid(row=4, column=1, sticky=N+S+E+W )
        self.mainIntVar = IntVar()
        self.mainMotor = Checkbutton( self, indicatoron=0, 
                        bd=2, variable = self.mainIntVar, text="Main Brush",
                        command = self.setMainMotor )
        self.mainMotor.grid(row=4, column=2, sticky=N+S+E+W )
        self.driveButton = Button(self, text="Drive mm/s mm", command=self.go)
        self.driveButton.grid(row=4, column=3, sticky=N+S+E+W )
        self.velentryv = IntVar()
        self.velentry = Entry(self, textvariable=self.velentryv)
	self.velentry.bind("<Return>", self.checkvelentry)
        self.velentry.grid(row=4,column=4,sticky=N+S+E+W)
        self.radentryv = IntVar()
        self.radentry = Entry(self, textvariable=self.radentryv)
	self.radentry.bind("<Return>", self.checkradentry)
        self.radentry.grid(row=4,column=5,sticky=N+S+E+W)

        # LEDs
        self.spotLEDButton = Button(self, text="Spot LED", command=self.toggleSpotLED)
        self.spotLEDButton.grid(row=5, column=0, sticky=N+S+E+W )
        self.cleanLEDButton = Button(self, text="Clean LED", command=self.toggleCleanLED)
        self.cleanLEDButton.grid(row=5, column=1, sticky=N+S+E+W )
        self.powerLEDButton = Button(self, text="Power LED", command=self.togglePowerLED)
        self.powerLEDButton.grid(row=5, column=2, sticky=N+S+E+W )
        self.statusLEDButton = Button(self, text="Status LED", command=self.toggleStatusLED)
        self.statusLEDButton.grid(row=5, column=3, sticky=N+S+E+W )
        self.ddLEDButton = Button(self, text="DirtDetect LED", command=self.toggleDDLED)
        self.ddLEDButton.grid(row=5, column=4, sticky=N+S+E+W )
        self.maxLEDButton = Button(self, text="Max LED", command=self.toggleMaxLED)
        self.maxLEDButton.grid(row=5, column=5, sticky=N+S+E+W )


        # Buttons
        self.spotButtonButton = Button(self, text="Spot Button", command=self.handleSpotButton)
        self.spotButtonButton.grid(row=6, column=0, sticky=N+S+E+W )
        self.cleanButtonButton = Button(self, text="Clean Button", command=self.handleCleanButton)
        self.cleanButtonButton.grid(row=6, column=1, sticky=N+S+E+W )
        self.powerButtonButton = Button(self, text="Power Button", command=self.handlePowerButton)
        self.powerButtonButton.grid(row=6, column=2, sticky=N+S+E+W )
        self.remoteButtonButton = Button(self, text="Remote Command", command=self.handleRemoteButton)
        self.remoteButtonButton.grid(row=6, column=3, sticky=N+S+E+W )
        self.remoteentryv = IntVar()
        self.remoteentry = Entry(self, textvariable=self.remoteentryv)
	self.remoteentry.bind("<Return>", self.checkremoteentry)
        self.remoteentry.grid(row=6,column=4,sticky=N+S+E+W)
        self.maxButtonButton = Button(self, text="Max Button", command=self.handleMaxButton)
        self.maxButtonButton.grid(row=6, column=5, sticky=N+S+E+W )

	# odometry line
	self.odometryButton = Button(self, text="Reset Odometry", height=2, command=self.resetOdometryButton)
	self.odometryButton.grid(row=15, column=0, rowspan=2, sticky=N+S+E+W )
	self.xLabel = Label(self, text="x in mm", height=1, fg="blue")
	self.xLabel.grid(row=15, column=1, sticky=N+S+E+W )
	self.xValue = Label(self, text="Unknown", height=1)
	self.xValue.grid(row=16, column=1, sticky=N+S+E+W )
	self.yLabel = Label(self, text="y in mm", height=1, fg="blue")
	self.yLabel.grid(row=15, column=2, sticky=N+S+E+W )
	self.yValue = Label(self, text="Unknown", height=1)
	self.yValue.grid(row=16, column=2, sticky=N+S+E+W )
	self.tLabel = Label(self, text="theta in deg", height=1, fg="blue")
	self.tLabel.grid(row=15, column=3, sticky=N+S+E+W )
	self.tValue = Label(self, text="Unknown", height=1)
	self.tValue.grid(row=16, column=3, sticky=N+S+E+W )
	self.rawDistLabel = Label(self, text="raw distance", height=1, fg="blue")
	self.rawDistLabel.grid(row=15, column=4, sticky=N+S+E+W )
	self.rawDistValue = Label(self, text="Unknown", height=1)
	self.rawDistValue.grid(row=16, column=4, sticky=N+S+E+W )
	self.rawAngleLabel = Label(self, text="raw angle", height=1, fg="blue")
	self.rawAngleLabel.grid(row=15, column=5, sticky=N+S+E+W )
	self.rawAngleValue = Label(self, text="Unknown", height=1)
	self.rawAngleValue.grid(row=16, column=5, sticky=N+S+E+W )


	# song facilities
	self.songLabel = Button(self, text="Play Song", height=1, fg="blue", command=self.songPlayOnly)
	self.songLabel.grid(row=10, column=0, sticky=N+S+E+W)
	self.songnumberv = IntVar()
	self.songNumberEntry = Entry(self, textvariable=self.songnumberv)
	self.songNumberEntry.bind("<Return>", self.checksongnumberentry)
        self.songNumberEntry.grid(row=11, column=0, sticky=N+S+E+W)
	self.songSetButton = Button(self, text="Song Set", height=2, command=self.handleSongSetButton)
	self.songSetButton.grid(row=10, column=1, rowspan=2, sticky=N+S+E+W)
	self.songPlayButton = Button(self, text="Set & Play", height=2, command=self.handleSongPlayButton)
	self.songPlayButton.grid(row=10, column=2, rowspan=2, sticky=N+S+E+W)
	self.songv = StringVar()
	self.songEntry = Entry(self, textvariable=self.songv)
        self.songv.set('60 16 64 16 67 16 72 32')
	self.songEntry.bind("<Return>", self.checksongentry)
        self.songEntry.grid(row=10, column=3, rowspan=2, columnspan=3, sticky=N+S+E+W)

        # canvases
        self.canvas1 = ErdosCanvas(self)
        self.canvas1.tfm.setScales(1,-1)
        self.canvas1.tfm.setWorldRotationCenter(0,0)
        self.canvas1.loadMap('mapColors.txt')
        self.robot = MovableCircle(self.canvas1,colorstr='blue')  
        #self.canvas1.bind("<Button-1>", self.handleCanvas1b1)

        self.canvas1.grid(row=20, column=0, columnspan=3, sticky=N+S+E+W)

        # not yet used...
        self.canvas2 = Canvas(self,bg="green",height=200)
        self.canvas2.bind("<Button-1>", self.handleCanvas1b1)
        self.canvas2.grid(row=20, column=3, columnspan=3, sticky=N+S+E+W)

	# electrical line
	self.chargingStateLabel = Label(self, text="Charging State", height=1, fg="blue")
	self.chargingStateLabel.grid(row=24, column=0, sticky=N+S+E+W )
	self.chargingState = Label(self, text="Unknown", height=1)
	self.chargingState.grid(row=25, column=0, sticky=N+S+E+W )
	self.voltageLabel = Label(self, text="Voltage mV", height=1, fg="blue")
	self.voltageLabel.grid(row=24, column=1, sticky=N+S+E+W )
	self.voltage = Label(self, text="Unknown", height=1)
	self.voltage.grid(row=25, column=1, sticky=N+S+E+W )
	self.currentLabel = Label(self, text="Current mA", height=1, fg="blue")
	self.currentLabel.grid(row=24, column=2, sticky=N+S+E+W )
	self.current = Label(self, text="Unknown", height=1)
	self.current.grid(row=25, column=2, sticky=N+S+E+W )
	self.chargeLabel = Label(self, text="Charge mAh", height=1, fg="blue")
	self.chargeLabel.grid(row=24, column=3, sticky=N+S+E+W )
	self.charge = Label(self, text="Unknown", height=1)
	self.charge.grid(row=25, column=3, sticky=N+S+E+W )
	self.capacityLabel = Label(self, text="Capacity mAh", height=1, fg="blue")
	self.capacityLabel.grid(row=24, column=4, sticky=N+S+E+W )
	self.capacity = Label(self, text="Unknown", height=1)
	self.capacity.grid(row=25, column=4, sticky=N+S+E+W )
	self.tempLabel = Label(self, text="Temperature C", height=1, fg="blue")
	self.tempLabel.grid(row=24, column=5, sticky=N+S+E+W )
	self.temp = Label(self, text="Unknown", height=1)
	self.temp.grid(row=25, column=5, sticky=N+S+E+W )


        # radiobuttons for the robot SCI state (mode)
        self.scistatev = IntVar()
        self.offRadio = Radiobutton(self,variable=self.scistatev,text='Off Mode',value=0,indicatoron=0,command=self.handleSCIMode)
        self.offRadio.grid(row=28, column=0, sticky=N+S+E+W)
        self.passiveRadio = Radiobutton(self,variable=self.scistatev,text='Passive Mode',value=1,indicatoron=0,command=self.handleSCIMode)
        self.passiveRadio.grid(row=28, column=1, sticky=N+S+E+W)
        self.safeRadio = Radiobutton(self,variable=self.scistatev,text='Safe Mode',value=2,indicatoron=0,command=self.handleSCIMode)
        self.safeRadio.grid(row=28, column=2, sticky=N+S+E+W)
        self.fullRadio = Radiobutton(self,variable=self.scistatev,text='Full Mode',value=3,indicatoron=0,command=self.handleSCIMode)
        self.fullRadio.grid(row=28, column=3, sticky=N+S+E+W)
        # power on buttons
        self.rooToothPowerButton = Button(self, text="RooTooth Power", command=self.handleRooToothPower)
        self.rooToothPowerButton.grid(row=28, column=4, sticky=N+S+E+W)
        self.rtsPowerButton = Button(self, text="RTS Power", command=self.handleRTSPower)
        self.rtsPowerButton.grid(row=28, column=5, sticky=N+S+E+W)


        # connection controls
        self.quitButton = Button(self, text="Quit", command=self.exitGracefully)
        self.quitButton.grid(row=30,column=0,sticky=N+S+E+W)
        self.comlabel = Button(self, text="COM Port", command=self.openComPort)
        self.comlabel.grid(row=30,column=1, sticky=N+S+E+W)
        self.comentryv = StringVar()
        self.comentry = Entry(self, textvariable=self.comentryv)
	self.comentry.bind("<Return>", self.openComPort)
        self.comentry.grid(row=30,column=2,sticky=N+S+E+W)

        self.serverlabel = Button(self, text="Server Port", command=self.openServerPort)
        self.serverlabel.grid(row=30, column=3, sticky=N+S+E+W)
        self.serverentryv = StringVar()
        self.serverentry = Entry(self, textvariable=self.serverentryv)
	self.serverentry.bind("<Return>", self.openServerPort)
        self.serverentry.grid(row=30,column=4,sticky=N+S+E+W)
        self.userButton = Button(self, text="User")
        self.userButton.grid(row=30,column=5,sticky=N+S+E+W)

    def handleCanvas1b1(self,ev):
        print 'in handleCanvas1b1'
        print 'ev x,y', ev.x, ev.y

    def handleRTSPower(self):
	""" assumes that it's a RooTooth, which has PIO pin 6
	    attached to the RTS serial line...
	"""
	if self.roomba != None:
            self.roomba.turnOnRTSPower()
        else:
            print 'No roomba connected!'

    def handleRooToothPower(self):
	""" assumes that it's a RooTooth, which has PIO pin 6
	    attached to the RTS serial line...
	"""
	if self.roomba != None:
            self.roomba.turnOnRooToothPower()
            # put into OFF Mode
            self.scistatev.set(0)
            # set the LEDs that seem to turn on...
            self.cleanLEDButton.configure(bg="green")
            self.spotLEDButton.configure(bg="green")
            self.powerLEDButton.configure(bg="green")
            # need to check about the internal LED state, perhaps...
        else:
            print 'No roomba connected!'

    def handleSCIMode(self):
	""" handleSCIMode changes the SCI mode (state) of the
	    roomba - since this is not directly queriable, it
	    makes its best guess at displaying things and
	    does its best to set the mode when asked

	    however, it's not possible to turn it to "off mode"
	    to turn it to "passive mode" we push the power button
	    (virtually)
	"""
        scimode = self.scistatev.get()
        ### print 'scimode is', scimode
	# both off mode and passive mode switch to passive mode
	# by pushing the (trying max) button (virtually)
        # 
        # OK, we assume the user doesn't _really_ want to be 
        # in off_mode, not to mention that I don't think 
        # it's even possible without taking out the battery...
        # so we simply use the OFF button as the power button
	if scimode == 0:  # pushed OFF_MODE
            self.roomba.buttonPress( roomba.POWER )
	elif scimode == 1:
	    # this actually seems to put it inot Passive Mode...
	    # this seems to be an undocumented feature
	    # like the virtual cliff sesnors on when it's charging
            self.roomba.start()  # send the "START" command
	    # this seems to turn on the spot, clean and power LEDs to green
	    # so, we'll reflect that in our gui...
            self.cleanLEDButton.configure(bg="green")
            self.spotLEDButton.configure(bg="green")
            self.powerLEDButton.configure(bg="green")
        elif scimode == 2: # want to go to safe mode...
            self.roomba.toSafeMode()
        elif scimode == 3: # want to go to full mode
            self.roomba.toFullMode()
            

    def resetOdometryButton(self):
        print 'in resetOdometryButton'

    def checksongentry(self,ev):
        """ this needs to be sure we have a legal song! 
            as such, it will be a bit more involved than the others
            a song must have
                0-16 notes, each from 30-127 inclusive (30 is a rest)
                alternating with durations from 0 to 255 inclusive
                space separated (not necessary, but reasonable...)
            this function returns a list of (freq,duration) tuples
            of the sort that Roomba.setSong requires
        """
        ss = self.songv.get() # get the string
        ssL = ss.split()      # split on the spaces
        finalL = []           # integer note list
        finalS = []           # string list for printing
        #print 'ssL is', ssL
        duration = False      # True if we're on a duration, not a note 
        for x in ssL[0:32]:   # ignore extras
            try:
                nextx = int(x)
                if not duration:
                    if nextx <  30: nextx =  30 # 30 == rest, 31 == lowest note
                    if nextx > 127: nextx = 127
                else:
                    if nextx <   0: nextx =   0
                    if nextx > 255: nextx = 255

                duration = not duration
                finalL.append(nextx)
                finalS = finalS + [ str(nextx), ' ' ]
            except:
                pass

        if duration == True:  # uh oh - we need a final duration
            finalL.append(64)     # one second
            finalS = finalS + [ '64', ' ' ]

        ss = ''.join(finalS)
        self.songv.set(ss)

        # finalL has to be a list of tuples for the setSong method
        i = 0
        tuplizedFinalL = []
        while i < len(finalL):
            tuplizedFinalL.append( (finalL[i],finalL[i+1]) )
            i += 2
        return tuplizedFinalL

    def handleSongPlayButton(self):
        """ handleSongPlayButton 
            first checks the song to make it legal
            then sets it to the current song number entry
            then plays that entry
        """
        songdatalist = self.checksongentry(None) # no event to pass
        self.checksongnumberentry(None)
        songnumber = self.songnumberv.get()
        # do we have a roomba?
        if self.roomba != None:
            self.roomba.setSong(songnumber, songdatalist)
            self.roomba.playSong(songnumber)
        else:
            print 'No roomba connected.'


    def songPlayOnly(self):
        """ plays the song whose number is shown
            (first checks for a legal number...)
        """
        self.checksongnumberentry(None)
        songnumber = self.songnumberv.get()
        if self.roomba != None:
            self.roomba.playSong(songnumber)
        else:
            print 'No roomba connected.'

    def handleSongSetButton(self):
        """ handleSongSetButton 
            first checks the song to make it legal
            then sets it to the current song number entry
        """
        songdatalist = self.checksongentry(None) # no event to pass
        self.checksongnumberentry(None)
        songnumber = self.songnumberv.get()
        # do we have a roomba?
        if self.roomba != None:
            self.roomba.setSong(songnumber, songdatalist)
        else:
            print 'No roomba connected.'

    def checksongnumberentry(self,ev):
        """ makes sure the song number entry is in the
            legal range of 0 to 15 inclusive
        """
        songnumber = self.songnumberv.get()
        if songnumber <  0: self.songnumberv.set( 0)
        if songnumber > 15: self.songnumberv.set(15)

    def handleSpotButton(self):
        """ a virtual button press of the "Spot" button
        """
        ### print 'in handleSpotButton'
        self.roomba.buttonPress( roomba.SPOT )
        self.scistatev.set(1) # put in PASSIVE_MODE

    def handleCleanButton(self):
        """ a virtual button press of the "Clean" button
        """
        self.roomba.buttonPress( roomba.CLEAN )
        self.scistatev.set(1) # put in PASSIVE_MODE

    def handlePowerButton(self):
        """ a virtual button press of the "Power" button
        """
        self.roomba.buttonPress( roomba.POWER )
        self.scistatev.set(0) # put in OFF_MODE

    def handleRemoteButton(self):
        """ a virtual button press of the remote
            this would be a good place to 
              - have the remote actually work as a remote
              - add one's own programs
        """
        print 'in handleRemoteButton'
        print 'self.remoteentryv is', self.remoteentryv.get()

    def checkremoteentry(self,ev):
        """ checkremoteentry makes sure the remote-control
            command is in the legal range of 0 to 255 inclusive
        """
        remoteentryvalue = self.remoteentryv.get()
        if remoteentryvalue <   0: self.remoteentryv.set(  0)
        if remoteentryvalue > 255: self.remoteentryv.set(255)

    def handleMaxButton(self):
        """ a virtual button press of the "Max" button
            note that the roomba red does not have a Max button
        """
        self.roomba.buttonPress( roomba.MAX )
        self.scistatev.set(1) # put in PASSIVE_MODE

    def toggleSpotLED(self):
        """ toggleSpotLED does just that
        """
        if self.roomba == None:
            print 'No roomba open'
        else:
            ledState = self.roomba.led[ roomba.LED_SPOT ]
            nextLEDState = 1-ledState
            self.roomba.setLED( roomba.LED_SPOT, nextLEDState )
            # set button background to the appropriate color
            if nextLEDState == 0: color = "lightgray"
            if nextLEDState == 1: color = "green"
            self.spotLEDButton.configure(bg=color)

    def toggleCleanLED(self):
        """ toggleCleanLED does just that
        """
        if self.roomba == None:
            print 'No roomba open'
        else:
            ledState = self.roomba.led[ roomba.LED_CLEAN ]
            nextLEDState = 1-ledState
            self.roomba.setLED( roomba.LED_CLEAN, nextLEDState )
            # set button background to the appropriate color
            if nextLEDState == 0: color = "lightgray"
            if nextLEDState == 1: color = "green"
            self.cleanLEDButton.configure(bg=color)

    def togglePowerLED(self):
        """ togglePowerLED does just that
        """
        if self.roomba == None:
            print 'No roomba open'
        else:
            ledState, ledOptState = self.roomba.led[ roomba.LED_POWER ]
	    if ledState < 128: # it's "off"
                nextLEDState = 255
                nextLEDOptState = 0
                color = "green"
            elif ledOptState < 100:
                nextLEDState = 255
                nextLEDOptState = 128
                color = "yellow"
            elif ledOptState < 200:
                nextLEDState = 255
                nextLEDOptState = 255
                color = "red"
            else:
                nextLEDState = 0
                nextLEDOptState = 0
                color = "lightgray"

            self.roomba.setLED( roomba.LED_POWER, nextLEDState, nextLEDOptState )
            # set button background to the appropriate color
            self.powerLEDButton.configure(bg=color)

    def toggleStatusLED(self):
        """ toggleStatusLED does just that...
        """
        if self.roomba == None:
            print 'No roomba open'
        else:
            ledState = self.roomba.led[ roomba.LED_STATUS ]
            nextLEDState = (ledState+1)%4
            self.roomba.setLED( roomba.LED_STATUS, nextLEDState )
            # set button background to the appropriate color
            if nextLEDState == 0: color = "lightgray"
            if nextLEDState == 1: color = "red"
            if nextLEDState == 2: color = "green"
            if nextLEDState == 3: color = "yellow"
            self.statusLEDButton.configure(bg=color)

    def toggleDDLED(self):
        """ toggleDDLED does just that...
        """
        if self.roomba == None:
            print 'No roomba open'
        else:
            ledState = self.roomba.led[ roomba.LED_DIRTDETECT ]
            nextLEDState = 1-ledState
            self.roomba.setLED( roomba.LED_DIRTDETECT, nextLEDState )
            # set button background to the appropriate color
            if nextLEDState == 0: color = "lightgray"
            if nextLEDState == 1: color = "blue"
            self.ddLEDButton.configure(bg=color)

    def toggleMaxLED(self):
        """ toggleMaxLED does just that...
            note that the Roomba Red does not have a Max LED
        """
        if self.roomba == None:
            print 'No roomba open'
        else:
            ledState = self.roomba.led[ roomba.LED_MAX ]
            nextLEDState = 1-ledState
            self.roomba.setLED( roomba.LED_MAX, nextLEDState )
            # set button background to the appropriate color
            if nextLEDState == 0: color = "lightgray"
            if nextLEDState == 1: color = "green"
            self.maxLEDButton.configure(bg=color)

    def setVacuumMotor(self):
        """ toggles the vacuum motor
            this is a good candidate for powering something in
            the back of the roomba's vacumming chamber
        """
        if self.roomba == None:
            print 'No roomba open'
        else:
            vacuumState = self.vacuumIntVar.get()
            self.roomba.setVacuum(vacuumState)
            print 'setting vacuum motor to', vacuumState

    def setSideMotor(self):
        """ toggles the side-brush motor
        """
        if self.roomba == None:
            print 'No roomba open'
        else:
            state = self.sideIntVar.get()
            self.roomba.setSideBrush(state)
            print 'setting side brush motor to', state

    def setMainMotor(self):
        """ toggles the main brushes' motor
        """
        if self.roomba == None:
            print 'No roomba open'
        else:
            state = self.mainIntVar.get()
            self.roomba.setMainBrush(state)
            print 'setting main brush motor to', state


    def openServerPort(self,ev):
        """ starts and stops (?) the server
        """
        print 'Open server port'
        port = self.serverentryv.get()
        # call correction here
        print 'desired port is', port
        self.threadedServerObject.setServerState(desiredServerRunning=True, 
                                                 PORT=5042, 
                                                 HOST='127.0.0.1') 

    def openComPort(self,ev=None):
        """ starts and stops (? not yet) the serial
            connection to the roomba
            this should really be done first and everything
            should be disabled until it's done... maybe someday
            also, this should allow a connection to a simulated roomba
        """
        comport = self.comentryv.get()
        if self.roomba == None:
            print 'Opening Roomba at COM Port', comport
            self.roomba = roomba.Roomba(int(comport))
            time.sleep(1)
            self.scistatev.set(0)
            print 'Roomba Open.'
        else:
            print 'Roomba already open!'

    def go(self):
        """ interface to the drive command
        """
        if self.roomba != None:
            self.checkradentry(None)
            self.checkvelentry(None)
            self.roomba.drive( self.velentryv.get(), self.radentryv.get() )
        else:
            print 'Need to connect to a roomba first'

    def checkradentry(self,ev):
        """ makes sure the radius of curvature is in bounds...
            also, you can use ROC==0 to specify straight
            1 and -1 specify turn-only (SCI specification)
        """
        rad = self.radentryv.get()
        if type(rad) != type(42):
            print 'radius must be an int in mm (-2000 to 2000)'
            print 'infinity (straight) is 32768'
            self.radentryv.set(32768)
        elif rad == 0:
            # special - make 0 "go straight"
            self.radentryv.set(32768)
        elif rad>2000:
            self.radentryv.set(32768)
        elif rad<-2000:
            self.radentryv.set(32768)

    def checkvelentry(self,ev):
        """ makes sure the velocity is in bounds...
        """
        vel = self.velentryv.get()
        if type(vel) != type(42):
            print 'velocity must be an int in mm/sec (-500 to 500)'
            self.velentryv.set(0)
        elif vel>500:
            self.velentryv.set(500)
        elif vel<-500:
            self.velentryv.set(-500)

    def stop(self):
        """ a shortcut for stopping 
        """
        if self.roomba != None:
            self.roomba.drive( 0, 0 )
            self.velentryv.set(0)
            self.radentryv.set(0)
        else:
            print 'Need to connect to roomba first!'

    def continuousUpdating(self):
        """ this toggles continuous updating of all sensor data
            see self.idleTasks for the scheduling of the
            individual frames' updating...
        """
        value = self.continuousupdatingv.get()
        if value == 1:
            print 'Turning on continuous updating...'
        else:
            print 'Turning off continuous updating...'

    def updateAll(self):
        """ this updates all three sensor frames
        """
        ### print 'Updating Everything'
        if self.roomba != None:
            self.sf3 = self.sf2 = self.sf = self.roomba.readSensors(0)
            if self.sf == None: return  # in case we're not in the right mode
            ### print self.sf
            self.visupdate1(self.sf)    # update the gui's components
            self.visupdate2(self.sf)    # for each of 
            self.visupdate3(self.sf)    # the sensor frames
        else:
            print 'Need to connect to roomba first!'

    def update1(self):
        """ update1 updates the data only from the 
            first sensor frame
        """
        if self.roomba != None:
            self.sf = self.roomba.readSensors(1)
            if self.sf == None: return  # in case we're not in the right mode
            self.visupdate1(self.sf)
        else:
            print 'Need to connect to roomba first!'

    def update2(self):
        """ update2 updates the data only from the 
            second sensor frame
        """
        if self.roomba != None:
            self.sf2 = self.roomba.readSensors(2)
            if self.sf2 == None: return  # in case we're not in the right mode
            self.visupdate2(self.sf2)
        else:
            print 'Need to connect to roomba first!'

    def update3(self):
        """ update3 updates the data only from the 
            third sensor frame
        """
        if self.roomba != None:
            self.sf3 = self.roomba.readSensors(3)
            if self.sf3 == None: return  # in case we're not in the right mode
            self.visupdate3(self.sf3)
        else:
            print 'Need to connect to roomba first!'

    def visupdate1(self, sf):
	""" update the GUI controls for sensor frame 1
	"""
        # left bump
        if sf.leftBump != 0:
            self.leftBumperLabel.configure(bg="red")
        else:
            self.leftBumperLabel.configure(bg="lightgray")
        # right bump
        if sf.rightBump != 0:
            self.rightBumperLabel.configure(bg="red")
        else:
            self.rightBumperLabel.configure(bg="lightgray")
        # wall
        if sf.wallSensor != 0:
            self.wallLabel.configure(bg="red")
        else:
            self.wallLabel.configure(bg="lightgray")
        # left wheel drop
        if sf.leftWheelDrop != 0:
            self.leftWheelDropLabel.configure(bg="red")
        else:
            self.leftWheelDropLabel.configure(bg="lightgray")
        # caster wheel drop
        if sf.casterDrop != 0:
            self.casterDropLabel.configure(bg="red")
        else:
            self.casterDropLabel.configure(bg="lightgray")
        # right wheel drop
        if sf.rightWheelDrop != 0:
            self.rightWheelDropLabel.configure(bg="red")
        else:
            self.rightWheelDropLabel.configure(bg="lightgray")
        # virtual wall
        if sf.virtualWall != 0:
            self.virtualWallLabel.configure(bg="red")
        else:
            self.virtualWallLabel.configure(bg="lightgray")
        # left cliff
        if sf.leftCliff != 0:
            self.leftCliffLabel.configure(bg="red")
        else:
            self.leftCliffLabel.configure(bg="lightgray")
        # left front cliff
        if sf.frontLeftCliff != 0:
            self.leftCentCliffLabel.configure(bg="red")
        else:
            self.leftCentCliffLabel.configure(bg="lightgray")
        # right front cliff
        if sf.frontRightCliff != 0:
            self.rightCentCliffLabel.configure(bg="red")
        else:
            self.rightCentCliffLabel.configure(bg="lightgray")
        # right cliff
        if sf.rightCliff != 0:
            self.rightCliffLabel.configure(bg="red")
        else:
            self.rightCliffLabel.configure(bg="lightgray")
	# overcurrents and dirt detect sensors
	if sf.driveLeft + sf.driveRight + sf.mainBrush +  \
           sf.vacuum + sf.sideBrush + sf.leftDirt + sf.rightDirt == 0:
            # everything is OK
            self.overcurrentsAndDDLabel.configure(bg="lightgray",text='OC/DD ok')
	else:
            ocddstr = 'OC/DD: ' + str(sf.driveLeft) + str(sf.driveRight) +  \
                      str(sf.mainBrush) + str(sf.vacuum) + str(sf.sideBrush) + \
                      str(sf.leftDirt) + str(sf.rightDirt)
            self.overcurrentsAndDDLabel.configure(bg="red",text=ocddstr)


    def visupdate2(self, sf):
	""" update the GUI controls for sensor frame 2
	"""
        # remote control command -- 255 is the default
        if sf.remoteControlCommand == 255:
            self.remoteButtonButton.configure(bg="lightgray")
        else:
            self.remoteButtonButton.configure(bg="red")
        # either way, put in the remote control signal seen...
        self.remoteentryv.set(sf.remoteControlCommand)
        # physical button presses by the user on the roomba itself
        if sf.spotButton == 1:
            self.spotButtonButton.configure(bg="red")
        else:
            self.spotButtonButton.configure(bg="lightgray")
        if sf.powerButton == 1:
            self.powerButtonButton.configure(bg="red")
        else:
            self.powerButtonButton.configure(bg="lightgray")
        if sf.cleanButton == 1:
            self.cleanButtonButton.configure(bg="red")
        else:
            self.cleanButtonButton.configure(bg="lightgray")
        if sf.maxButton == 1:
            self.maxButtonButton.configure(bg="red")
        else:
            self.maxButtonButton.configure(bg="lightgray")
        # raw distance and angle
        self.rawDistValue.configure(text=str(sf.distance))
        self.rawAngleValue.configure(text=str(sf.rawAngle))
        # odometric pose
        if self.roomba != None:
            xPose, yPose, thrPose = self.roomba.getPose()
            self.xValue.configure(text=('%.3f'%(xPose,)))
            self.yValue.configure(text=('%.3f'%(yPose,)))
            self.tValue.configure(text=('%.3f'%(math.degrees(thrPose),)))
        else:
            self.xValue.configure(text='No roomba')
            self.yValue.configure(text='No roomba')
            self.tValue.configure(text='No roomba')

    def visupdate3(self, sf):
	""" update the GUI controls for sensor frame 3
	"""
        # charging state
        chargingString = 'Not Charging'
        if sf.chargingState == 1: chargingString = 'Charging Recovery'
        elif sf.chargingState == 2: chargingString = 'Charging'
        elif sf.chargingState == 3: chargingString = 'Trickle Charging'
        elif sf.chargingState == 4: chargingString = 'Waiting'
        elif sf.chargingState == 5: chargingString = 'Charging Error'
        self.chargingState.configure(text=chargingString)
        # voltage
        self.voltage.configure(text=str(sf.voltage))
        # current
        self.current.configure(text=str(sf.current))
        # temperature
        self.temp.configure(text=str(sf.temperature))
        # charge
        self.charge.configure(text=str(sf.charge))
        # capacity
        self.capacity.configure(text=str(sf.capacity))

    def exitGracefully(self):
        print 'Exiting "gracefully"...'
        if self.roomba != None:
            self.roomba.close()
        self.quit()  # built-in quit function


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ #
#
# this ThreadedServer stuff can be removed if you want to
# use the gui alone
#
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ #

import threading
import socket

class ThreadedServer:

    def __init__(self, app):
        """ constructor for ThreadedServer
              starts the thread
              does not start the server
        """
        self.app = app        # the gui application

        self.s = None
        self.HOST = None
        self.PORT = None
        self.serverRunning = False
        self.maxMessageSize = 4096

        self.qToGUI = Queue.Queue()
        self.qFromGUI = Queue.Queue()
        self.app.setQueues(self.qToGUI, self.qFromGUI, self)

        self.serverthread = threading.Thread(target=self.server)
        self.serverthread.setDaemon(True)  # will quit if the gui does
        self.serverthread.start()

    def setServerState(self, desiredServerRunning, PORT=5042, HOST='127.0.0.1'):
        """ turns on/off the server
              input desiredServerRunning:  True to turn on
                                          False to turn off
        """
        if desiredServerRunning == self.serverRunning:
            print 'The server is already in the desired state,', desiredServerRunning
            return
        
        if desiredServerRunning == True:
            self.HOST = HOST
            self.PORT = PORT
            self.serverRunning = True
            # we let the server thread open things up...

        else:  # if desiredServerState == False
            self.serverRunning = False
            # we let the server thread close things down...


    # this is running in a separate thread!
    # think of it as another program
    #
    # the communication with the GUI is done through this
    def server(self):
        """ this runs the server (if turned on)
            it should allow for starting/stopping the server
            it should be graceful in handling disconnections...
        """
        # main loop
        while True:

            # connection-making loop
            while self.serverRunning == True:
                # create a new socket... old one will be collected
                self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.s.bind((self.HOST, self.PORT))
                self.s.listen(1)
                self.s.settimeout(0.1)
                print 'Listening for a connection...'
                networking = 0
                
                # loop to create a socket connection
                while networking == 0:
                    try:
                        # this will block for a little bit of time...
                        # see the timeout parameter above
                        conn, addr = self.s.accept()
                        networking = 1
                        print 'Connected by', addr
                    except socket.error:  # that is, no one connected...
                        networking = 0
                        time.sleep(0.5)
                        print 'No connection yet...'

                # loop for a single, established socket connection
                while networking == 1:
                
                    try:
                        data = conn.recv(self.maxMessageSize)  
                        if len(data) >= 1:
                            # question: how does the timeout affect conn?

                            # now handle the message here...
                            print 'Got data', data
                            retmsg = 'hi 42' # default return message

			    # if the message is a request for data,
			    # we simply send the latest data back
			    # and we don't bother the robot with it
			    if data == 'getPose':
				retmsg = '42.0 42.0 42.0'
		            elif data == 'quit': 
                                networking = 0
                                conn.close()
                            elif data[0:1] == 'h':
			        retmsg = str(self.app.serverentryv.get())
                            else:    # if the message is something else
                                self.qToGUI.put(data)
    
                                # generates more messages than the client
                                # is generating -> would create a backlog in 
                                # qFromGUI
                                try:
                                    retmsg = str(self.qFromGUI.get(block=False))
                                    print 'Sending back the msg', msg
                                except Queue.Empty:
                                    retmsg = 'hi 42'
    
                            # actually send the message back...
                            conn.send(retmsg)  
    
			    # I forget why this pause is important...
                            time.sleep(0.05)

                        else: # the len(data) == 0, i.e., data == ''
                            networking = 0
                            # self.qToGUI.put('reset')  # ?
                    except: # no data this time...
                        pass
                    # end of while networking == 1 loop

                print 'Connection broken'
                self.s.close()
                self.s = None   # will get reset at the top of the loop
                # end of while self.serverRunning == True loop

            ### print 'Server is no longer running'
            # sleep for a bit - see if the user
            # turns on the server...
            time.sleep(0.5)

            # end of while True loop
    





# if name == '__main__':

app = StateSlate()            # create the gui
server = ThreadedServer(app)  # a separate object for the server
# the server's loop will run in a separate thread
# To Do (?): allow the server to handle multiple connections...
#    this will likely require spawning a new thread (from within
#    the ThreadedServer object) for each connection... (?)

# this program's main loop is the gui's
# Tkinter likes to be the lead dog as far as I can tell...
app.master.title("ERDOS: Slate State Slate") 
app.mainloop()

